#include "scene.h"

#include "V3.h"
#include "M33.h"
#include "ppc.h"
#include "tm.h"
#include "IFB.h"

Scene *scene;

using namespace std;

#include <iostream>
#include <fstream>

Scene::Scene() {

	hwfb = 0;
	oldvisfb = 0;
	visfb = 0;
	visppc = 0;
	ifb = 0;


	lfb = 0;
	rfb = 0;
	panofb = 0;
	panofb0 = 0;
	lppc = rppc = panoppc = 0;

	gui = new GUI();
	gui->show();

	int u0 = 10;
	int v0 = 40;
	int h = 400;
	int w = 600;
	fb = new FrameBuffer(u0, v0, w, h);
	fb->position(u0, v0);
	fb->label("first");
	fb->show();
	fb->SetBGR(0xFFFFFFFF);
	fb->redraw();

	gui->uiw->position(u0, v0 + fb->h + v0);

	float hfov = 90.0f;
	ppc = new PPC(hfov, fb->w, fb->h);
	tmsN = 4;
	tms = new TM[tmsN];
	tms[0].LoadBin("geometry/teapot1K.bin");
//	tms[0].LoadBin("geometry/teapot57K.bin");
	V3 tv(20.0f, -20.0f, -200.0f);
	tms[0].Translate(tv);
//	tms[0].on = 0;



	FrameBuffer *texture = new FrameBuffer(10, 10, 128, 128);
	texture->SetAsChecker(0xFF000000, 0xFFAAAAAA, 32);
	texture->label("Texture");
//	texture->show();

	tms[1].SetQuad(V3(-100.0f, 100.0f, -350.0f),
		V3(-100.0f, -100.0f, -350.0f),
		V3(100.0f, -100.0f, -350.0f),
		V3(100.0f, 100.0f, -350.0f));
	tms[1].texture = texture;
	tms[1].on = 0;


	AABB aabb(V3(0.0f, 0.0f, 0.0f));
	aabb = tms[0].GetAABB();
	tms[2].SetQuad(
		V3(aabb.corners[0][0], aabb.corners[0][1], aabb.corners[0][2]),
		V3(aabb.corners[0][0], aabb.corners[0][1], aabb.corners[1][2]),
		V3(aabb.corners[1][0], aabb.corners[0][1], aabb.corners[1][2]),
		V3(aabb.corners[1][0], aabb.corners[0][1], aabb.corners[0][2])
		);
	V3 qc = tms[2].GetCenter();
	tms[2].Translate(qc * -1.0f);
	tms[2].Scale(3.0f);
	tms[2].Translate(qc);
	tms[2].SetAllColors(V3(0.8f, 1.0f, 0.8f));

	tms[3].SetQuad(
		V3(aabb.corners[0][0], aabb.corners[0][1], aabb.corners[0][2]),
		V3(aabb.corners[0][0], aabb.corners[0][1], aabb.corners[1][2]),
		V3(aabb.corners[0][0], aabb.corners[1][1], aabb.corners[1][2]),
		V3(aabb.corners[0][0], aabb.corners[1][1], aabb.corners[0][2])
	);
	tms[3].Rotate(tms[3].GetCenter(), V3(1.0f, 0.0f, 0.0f), 5.0f);
	tms[3].Translate(V3(-15.0f, -10.0f, 0.0f));

	tms[3].SetDisk(300);
	tms[3].Scale(30.0f);
	tms[3].Rotate(V3(0.0f, 0.0f, 0.0f), V3(0.0f, 1.0f, 0.0f), 90.0f);
	tms[3].Translate(tms[0].GetCenter());
	tms[3].Translate(V3(-60.0f, -35.0f, 0.0f));

	tms[3].SetAllColors(V3(0.8f, 0.8f, 1.0f));


	fb3 = new FrameBuffer(u0, v0, w, h);
	fb3->position(u0 + w + u0, v0);
	fb3->label("third");
	fb3->show();
	fb3->SetBGR(0xFFFFFFFF);
	fb3->redraw();
	ppc3 = new PPC(45.0f, fb3->w, fb3->h);
	ppc3->SetPose(ppc3->C + V3(100.0f, 200.0f, 400.0f), tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));


	L = ppc->C;
	L = tms[0].GetCenter() + V3(-100.0f, 90.0f, -50.0f);
	ka = 0.5f;
	shfb = 0;
	shppc = 0;
	shadowsOn = 0;

}

void Scene::Render() {

	if (panofb) {
//		fb->SetFromEQRI(ppc, panofb);
		fb->SetFromFisheye(ppc, panofb);
		fb->redraw();
		return;
	}

	Render(ppc, fb);

	if (hwfb) {
		hwfb->redraw();
	}

	return;

	tms[2].on = 0;
	Render(ppc3, fb3);
	tms[2].on = 1;

	fb3->Draw3DPoint(L, V3(0.9f, 0.7f, 0.0f), 17, ppc3);

	float visFocalLength = 20.0f;
	ppc->Visualize(visFocalLength, ppc3, fb3);

	fb->VisualizeSamples(ppc, ppc3, fb3);

//	fb->VisualizeSamples(ppc, ppc3, fb3, ppc->GetF() / visFocalLength);

//	tms[0].VisualizeNormals(5.0f, ppc3, fb3);

}

void Scene::RenderHW(PPC *cppc, FrameBuffer *cfb) {

//	cfb->SetBGR(0xFF880000);

	glClearColor(0.75f, 0.5f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	ppc->SetIntrinsicsHW();
	ppc->SetExtrinsicsHW();

	for (int tmi = 0; tmi < tmsN; tmi++) {
		if (!tms[tmi].on)
			continue;
		tms[tmi].RenderHW();
	}

}


void Scene::Render(PPC *cppc, FrameBuffer *cfb) {

	cfb->SetBGR(0xFF880000);
	cfb->ClearZB();
	int psize = 17;
	for (int tmi = 0; tmi < tmsN; tmi++) {
		if (!tms[tmi].on)
			continue;
//		if (cppc == ppc3)
//			continue;
//		tms[tmi].RenderPoints(cfb, psize, cppc);
//		tms[tmi].RenderWF(cfb, cppc);
		tms[tmi].RenderFilled(cfb, cppc);
	}
	cfb->redraw();
}

void Scene::DBG() {

	{
		fb3->hide();
		int u0 = 10;
		int v0 = 40;
		int h = 400;
		int w = 600;
		hwfb = new FrameBuffer(u0 + fb->w + u0, v0, w, h);
		hwfb->label("hw");
		hwfb->show();
		hwfb->ishw = 1;
		Render();
		return;
	}

	{
		panofb = new FrameBuffer(40 + fb->w + 40, 40, 10, 10);
//		panofb->LoadTiff("mydbg/images/cubeFisheye.tif");
		panofb->LoadTiff("mydbg/images/fisheye.tif");
		panofb0 = new FrameBuffer(40 + fb->w + 40, 40, 10, 10);
//		panofb0->LoadTiff("mydbg/images/cubeFisheye.tif");
		panofb0->LoadTiff("mydbg/images/fisheye.tif");
		cerr << "INFO: pano res: " << panofb->w << " " << panofb->h << endl;
		panofb->label("fisheye");
		panofb->show();
		fb3->hide();
		Render();
		return;
	}



	{
		panofb = new FrameBuffer(40 + fb->w + 40, 40, 10, 10);
//		panofb->LoadTiff("mydbg/images/panoSmall.tif");
		panofb->LoadTiff("mydbg/images/pano.tif");
		panofb0 = new FrameBuffer(40 + fb->w + 40, 40, 10, 10);
//		panofb0->LoadTiff("mydbg/images/panoSmall.tif");
		panofb0->LoadTiff("mydbg/images/pano.tif");
		cerr << "INFO: pano res: " << panofb->w << " " << panofb->h << endl;
		panofb->label("equirectangular");
		panofb->show();
		fb3->hide();
		Render();
		return;
	}

	{

		TM tm;
		char fname[] = "C:/Users/Voicu Popescu/Downloads/rp_aliyah_4d_004_dancing_4k.bins";
		ifstream ifs(fname, ios::binary);
		if (ifs.fail()) {
			cerr << "INFO: cannot open " << fname << endl;
		}
		unsigned int framesN;
		ifs.read((char*)&framesN, sizeof(unsigned int));
		cerr << "INFO: framesN = " << framesN << endl;
		ofstream ofs("mydbg/t100.txt");
		for (unsigned int fi = 0; fi < framesN; fi++) {
			unsigned int vertsN;
			ifs.read((char*)&vertsN, sizeof(unsigned int));
			cerr << "INFO: vertsN = " << vertsN << endl;

			unsigned char xyzFlag;
			ifs.read((char*)&xyzFlag, sizeof(unsigned char));
			cerr << "INFO: xyzFlag = " << xyzFlag << endl;
			unsigned char colorsFlag;
			ifs.read((char*)&colorsFlag, sizeof(unsigned char));
			cerr << "INFO: colorsFlag = " << colorsFlag << endl;
			unsigned char normalsFlag;
			ifs.read((char*)&normalsFlag, sizeof(unsigned char));
			cerr << "INFO: normalsFlag = " << normalsFlag << endl;
			unsigned char tcsFlag;
			ifs.read((char*)&tcsFlag, sizeof(unsigned char));
			cerr << "INFO: tcsFlag = " << tcsFlag << endl;

			tm.Allocate(vertsN, 1);
			ifs.read((char*)tm.verts, vertsN * sizeof(V3));

			AABB aabb = tm.GetAABB();
			cerr << "AABB: ";
			cerr << aabb.corners[0] << "; " << aabb.corners[1] << endl;

			if (colorsFlag == 'y') {
				ifs.read((char*)tm.colors, vertsN * sizeof(V3));
			}
			if (normalsFlag == 'y') {
				ifs.read((char*)tm.normals, vertsN * sizeof(V3));
			}
			if (tcsFlag == 'y') {
				float *tcs = new float[2 * vertsN];
				ifs.read((char*)tcs, vertsN * sizeof(float) * 2);
				delete[]tcs;
			}

			unsigned int trisN;
			ifs.read((char*)&trisN, sizeof(unsigned int));
			cerr << "INFO: trisN = " << trisN << endl;
			tm.trisN = trisN;
			tm.tris = new unsigned int[3 * trisN];
			ifs.read((char*)tm.tris, trisN * 3 * sizeof(unsigned int));

			if (fi == 0) {
				ppc->SetPose(tm.GetCenter() + V3(0.0f, 0.0f, 500.0f), tm.GetCenter(), V3(0.0f, 1.0f, 0.0f));
			}
			tm.SetAllColors(V3(1.0f, 0.0f, 0.0f));
			tm.Light(ppc->C, 0.1f);
			fb->ClearZB();
			fb->SetBGR(0xFFFFFFFF);
			tm.RenderPoints(fb, 1, ppc);
//			tm.RenderFilled(fb, ppc);
//			tm.RenderWF(fb, ppc);
			fb->redraw();
			Fl::check();
			ofs << tm.tris[3*100+0] << " " 
				<< tm.tris[3 * 100 + 1] << " "
				<< tm.tris[3 * 100 + 2]
				<< endl;
		}
		ofs.close();
		ifs.close();

		return;

		framesN = 1000;
		AABB aabb = tm.GetAABB();
		V3 mc = aabb.corners[0]*0.5f + aabb.corners[1]*0.5f;
		ppc->SetPose(mc + V3(0.0f, 0.0f, 300.0f), mc, V3(0.0f, 1.0f, 0.0f));
		for (unsigned int fi = 0; fi < framesN; fi++) {
			ppc->C = ppc->C.RotateThisPointAboutAxis(mc, V3(0.0f, 1.0f, 0.0f),
				360.0f / (float)(framesN - 1));
			ppc->SetPose(ppc->C, mc, V3(0.0f, 1.0f, 0.0f));
		
			fb->ClearZB();
			fb->SetBGR(0xFFFFFFFF);
			//			tm.RenderFilled(fb, ppc);
			tm.RenderWF(fb, ppc);
			fb->redraw();
			Fl::check();

		}

		return;

	}

	{

		fb->hide();
		fb3->hide();
		lfb = new FrameBuffer(10, 30, 10, 10);
		lfb->LoadTiff("mydbg/Images/LeftVSmall.tif");
		//		lfb->SaveAsTiff("mydbg/Images/LeftVSmall.tif");
		lfb->label("left");
		lfb->show();
		rfb = new FrameBuffer(10 + lfb->w + 10, 30, 10, 10);
		rfb->LoadTiff("mydbg/Images/RightVSmall.tif");
		//		rfb->SaveAsTiff("mydbg/Images/RightVSmall.tif");
		rfb->label("right");
		rfb->show();
		int panow = 800;
		int panoh = 400;
		panofb = new FrameBuffer(300, 30 + lfb->h + 60, panow, panoh);
		panofb->label("pano");
		panofb->show();

		float hfov = 55.0f;
		lppc = new PPC(hfov, lfb->w, lfb->h);
		rppc = new PPC(hfov, rfb->w, rfb->h);
		panoppc = new PPC(120.0f, panofb->w, panofb->h);

		float minErr = FLT_MAX;
		V3 ss0(60.0f, -30.0f, -5.0f);
		V3 ss1(90.0f, -20.0f, +5.0f);
		V3 bestGuess;
		int guessesN = 100000;
		for (int i = 0; i < guessesN; i++) {

			V3 guess; guess = guess.GetRandom(ss0, ss1);
			*lppc = PPC(guess[0], lppc->w, lppc->h);
			*rppc = *lppc;
			rppc->Pan(guess[1]);
			rppc->Tilt(guess[2]);

			float currErr = lfb->ColorDifference(rfb, rppc, lppc);
			if (currErr < minErr) {
				cerr << endl << "INFO: currerr, minErr:" << currErr << " " << minErr << endl;
				cerr << "INFO: curr guess:" << guess << endl;
				minErr = currErr;
				bestGuess = guess;

				panofb->SetBGR(0xFF808080);
				panofb->SetFromImage(lfb, lppc, panoppc);
				panofb->SetFromImage(rfb, rppc, panoppc);
				panofb->redraw();
				Fl::check();

			}

			if ((i + 1) % 1000 == 0) {
				cerr << "INFO: " << 100.0f * (float)i / (float)guessesN << "          \r";
			}

		}

		cerr << endl << "Done" << endl;
		cerr << "INFO: best guess: " << bestGuess << endl;

		return;



	}

	{
		tms[0].on = 0;
		shfb = new FrameBuffer(10, 10 + fb->h + 65, 256, 256);
		shppc = new PPC(45.0f, shfb->w, shfb->h);
		V3 gpc = tms[2].GetCenter();
		ppc->SetPose(V3(gpc[0], gpc[1]+50.0f, gpc[2]+200.0f), gpc, V3(0.0f, 1.0f, 0.0f));

		shfb->show();
		shfb->label("shadow view");
		L = tms[3].GetCenter() + V3(-150.0f, 50.0f, 0.0f);

		L = L + V3(0.0f, 0.2f, 0.0f);
		shppc->SetPose(L, tms[2].GetCenter(), V3(0.0f, 1.0f, 0.0f));


		ifb = new IFB(shppc->w, shppc->h);
		visfb = new FrameBuffer(10 + shfb->w + 50, 10 + fb->h + 65, ifb->w, ifb->h);
		oldvisfb = new FrameBuffer(10 + shfb->w + 50, 10 + fb->h + 65, ifb->w, ifb->h);


		visfb = new FrameBuffer(10 + shfb->w + 50, 10 + fb->h + 65, ifb->w, ifb->h);
		oldvisfb = new FrameBuffer(10 + shfb->w + 50, 10 + fb->h + 65, ifb->w, ifb->h);
		FrameBuffer *asfb = new FrameBuffer(10 + fb->w + 20, 40, fb->w, fb->h);
		asfb->label("Pixel accurate shadows");
		asfb->show();
		visfb->label("IFB vis");
		visfb->show();

		ppc->SetPose(ppc->C, V3(103.f, -20.f, -185.f), V3(0.0f, 1.0f, 0.0f));

		for (int fi = 0; fi < 1000; fi++) {

			shadowsOn = 0;
			Render(shppc, shfb);
			shadowsOn = 1;

			Render();
			ifb->Clear();

			ifb->AddSamplingLocations(shppc, fb, ppc);

			//		ifb->Visualize(visfb, V3(0.0f, 0.0f, 0.0f), V3(1.0f, 0.0f, 0.0f), 0, 7);
			ifb->Visualize(visfb, V3(0.0f, 0.0f, 0.0f), V3(1.0f, 0.0f, 0.0f));
			oldvisfb->CopyImage(visfb);
//			cerr << ifb->Max() << endl;
			visppc = shppc;
			int oldShadowsOn = scene->shadowsOn;
			scene->shadowsOn = 0;
			Render(ppc, asfb);
			scene->shadowsOn = oldShadowsOn;
			tms[3].PixelAccurateShadows(ifb, shppc, asfb);

//			L = L + V3(0.0f, .05f, 0.0f);
//			shppc->C = L;

			ppc->c = ppc->c + ppc->GetVD()*5.0f;

			Fl::check();
//			break;
		}

		return;
	}

	{
		tms[1].texture->hide();
		fb3->hide();
		for (int fi = 0; fi < 1000; fi++) {
			tms[1].Rotate(tms[1].GetCenter(), V3(0.0f, 1.0f, 0.0f), 0.05f);
			Render();
			Fl::check();
		}
		return;
	}

	{
		Render();
		return;
	}


	{
		ka = 0.2f;
		int framesN = 1000;
		V3 L0 = ppc->C; V3 L1 = tms[0].GetCenter() + V3(0.0f, 0.0f, -50.0f);
		for (int fi = 0; fi < framesN; fi++) {
			L = L0 + (L1 - L0)*(float)fi / (float)(framesN - 1);
			tms[0].SetAllColors(V3(1.0f, 0.0f, 0.0f));
			tms[0].Light(L, ka);
			Render();
			Fl::check();
		}
		return;

	}

	{
		int framesN = 720;
		V3 tmc = tms[0].GetCenter();
		PPC ppc0(*ppc);
		PPC ppc1(*ppc);
		V3 LaP = tmc;
		V3 newC = LaP + V3(0.0f, 100.0f, 100.0f);
		ppc1.SetPose(newC, LaP, V3(0.0f, 1.0f, 0.0f));

		for (int fi = 0; fi < framesN; fi++) {
			ppc->SetInterpolated(&ppc0, &ppc1, fi, framesN);
			Render();
			Fl::check();
		}
		*ppc = ppc0;
		return;
	}

	{

		V3 tmc = tms[0].GetCenter();
		V3 LaP = tmc;
		V3 newC = LaP + V3(0.0f, 200.0f, 0.0f);
		ppc->SetPose(newC, LaP, V3(0.0f, 0.0f, -1.0f));
		Render();
		return;

	}

	{
		int framesN = 720;
		V3 tmc = tms[0].GetCenter();
		for (int fi = 0; fi < framesN; fi++) {
//			tms[0].Translate(V3(0.1f, 0.0f, 0.0f));
			tms[0].Rotate(tmc, V3(0.0f, 1.0f, 0.0f), 1.0f);
//			ppc->Roll(1.f);
			Render();
			Fl::check();
		}
		return;
	}

	{
		float hfov = 55.0f;
		PPC ppc(hfov, fb->w, fb->h);

		TM tm;
		tm.LoadBin("geometry/teapot1K.bin");
//		tm.LoadBin("geometry/teapot57K.bin");
		V3 tv(20.0f, -20.0f, -200.0f);
		tm.Translate(tv);

		fb->SetBGR(0xFFFFFFFF);
		fb->ClearZB();

		tm.RenderPoints(fb, 3, &ppc);
		fb->redraw();
		return;

		V3 q[4];
		float hw = 10.0f;
		float z = -40.0f;
		q[0] = V3(-hw, +hw, z);
		q[1] = V3(-hw, -hw, z);
		q[2] = V3(+hw, -hw, z);
		q[3] = V3(+hw, +hw, z);

		V3 pP;
		ppc.Project(q[0], pP);
		cerr << pP << endl;
		ppc.Project(q[1], pP);
		cerr << pP << endl;
		ppc.Project(q[2], pP);
		cerr << pP << endl;
		ppc.Project(q[3], pP);
		cerr << pP << endl;

		TM quad;
		quad.SetQuad(q[0], q[1], q[2], q[3]);

		fb->SetBGR(0xFFFFFFFF);
		for (int vi = 0; vi < quad.vertsN; vi++) {
			fb->Draw3DPoint(quad.verts[vi], V3(0.0f, 0.0f, 0.0f), 7, &ppc);
		}
		fb->redraw();
		return;
	}

	cerr << "INFO: pressed DBG" << endl;
}

void Scene::NewButton() {
	cerr << "INFO: pressed New button on GUI" << endl;
}




